#include <iostream>
#include <sstream>
#include <cassert>

#include "gmsh_io.h"
#include "entity.h"

std::string
quadrature_name(int order)
{
    std::stringstream ss;
    ss << "Gauss" << order;
    return ss.str();
}

std::string
basis_func_name(int order)
{
    (void) order;
    std::stringstream ss;
    //ss << "Lagrange";
    ss << "H1Legendre" << order;
    return ss.str();
}

std::string
basis_grad_name(int order)
{
    (void) order;
    std::stringstream ss;
    //ss << "GradLagrange";
    ss << "GradH1Legendre" << order;
    return ss.str();
}

model::model()
    : geometric_order(1), approximation_order(1)
{
    remesh(geometric_order);
}

model::model(int pgo, int pao)
    : geometric_order(pgo), approximation_order(pao)
{
    assert(geometric_order > 0);
    assert(approximation_order > 0);

    remesh(geometric_order);
}

model::model(const char *filename)
    : geometric_order(1), approximation_order(1)
{
    gmsh::open( std::string(filename) );
    remesh(geometric_order);
}

model::model(const char *filename, int pgo, int pao)
    : geometric_order(pgo), approximation_order(pao)
{
    assert(geometric_order > 0);
    assert(approximation_order > 0);

    gmsh::open( std::string(filename) );
    remesh(geometric_order);
}

model::~model()
{
    gmsh::clear();
}

void
model::remesh()
{
    gmm::generate( DIMENSION(3) );
    gmm::setOrder( geometric_order );

    gvp_t gmsh_entities;

    /* Create a table mapping the tag of a boundary to the list of
     * its face keys. This must happen before import_gmsh_entities()
     * because otherwise you get spurious 2D entities (and this in
     * turn because the constructor of entity calls addDiscreteEntity()
     * to get all the element faces) .*/
    std::map<size_t, std::vector<face_key>>   boundaries_elemTags;
    gm::getEntities(gmsh_entities, DIMENSION(2));
    for (auto [dim, tag] : gmsh_entities)
    {
        std::vector<int> elemTypes;
        gmm::getElementTypes(elemTypes, dim, tag);
        assert(elemTypes.size() == 1);
        for (auto& elemType : elemTypes)
        {
            element_key_factory ekf(DIMENSION(2), tag, elemType);
            boundary_map[tag] = ekf.get_keys();
        }
    }

    entities.clear();
    import_gmsh_entities();
    map_boundaries(); /* This must happen after import_gmsh_entities(). */
}

void
model::remesh(int pgo)
{
    geometric_order = pgo;
    remesh();
}

std::vector<entity>::const_iterator
model::begin() const
{
    return entities.begin();
}

std::vector<entity>::const_iterator
model::end() const
{
    return entities.end();
}

std::vector<entity>::iterator
model::begin()
{
    return entities.begin();
}

std::vector<entity>::iterator
model::end()
{
    return entities.end();
}

entity&
model::entity_at(size_t pos)
{
    return entities.at(pos);
}

const entity&
model::entity_at(size_t pos) const
{
    return entities.at(pos);
}

size_t
model::num_dofs(void) const
{
    size_t ret = 0;
    for (const auto& e : entities)
        ret += e.num_dofs();

    return ret;
}

size_t
model::num_fluxes(void) const
{
    size_t ret = 0;
    for (const auto& e : entities)
        ret += e.num_fluxes();

    return ret;
}

size_t
model::num_cells(void) const
{
    size_t ret = 0;
    for (const auto& e : entities)
        ret += e.num_cells();

    return ret;
}

size_t
model::num_faces(void) const
{
    size_t ret = 0;
    for (const auto& e : entities)
        ret += e.num_faces();

    return ret;
}

size_t
model::num_entities(void) const
{
    return entities.size();
}

void
model::update_connectivity(const entity& e, size_t entity_index)
{
    for (size_t iT = 0; iT < e.num_cells(); iT++)
    {
        for (size_t iF = 0; iF < e.num_faces_per_elem(); iF++)
        {
            size_t findex = e.num_faces_per_elem() * iT  + iF;
            neighbour_descriptor cd(entity_index, iT, iF);
            element_key fk( e.face(findex) );
            conn.connect( fk, cd );
        }
    }
}

void
model::import_gmsh_entities(void)
{
    gvp_t gmsh_entities;
    gm::getEntities(gmsh_entities, DIMENSION(3));

    size_t entity_index = 0;
    size_t dof_base = 0;
    size_t flux_base = 0;
    size_t index_base = 0;
    for (auto [dim, tag] : gmsh_entities)
    {
        std::vector<int> elemTypes;
        gmm::getElementTypes(elemTypes, dim, tag);
        
        assert(elemTypes.size() == 1);
        if (elemTypes.size() != 1)
            throw std::invalid_argument("Only one element type per entity is allowed");

        for (auto& elemType : elemTypes)
        {
            entity e({.dim = dim, .tag = tag, .etype = elemType,
                      .gorder = geometric_order, .aorder = approximation_order});
            
            e.sort_by_orientation();
            e.base(dof_base, flux_base, index_base);
            e.number(entity_index);
            dof_base += e.num_dofs();
            flux_base += e.num_fluxes();
            index_base += e.num_cells();

            update_connectivity(e, entity_index);

            entities.push_back( std::move(e) );
            entity_index++;
        }
    }
}

void
model::map_boundaries(void)
{
    /* Make a vector mapping element_key to entity tag */
    using bfk_t = std::pair<element_key, int>;
    std::vector<bfk_t> bfk;
    for (auto& [tag, keys] : boundary_map)
        for (auto& k : keys)
            bfk.push_back( std::make_pair(k, tag) );

    /* Sort it for lookups */
    auto comp = [](const bfk_t& a, const bfk_t& b) -> bool {
        return a.first < b.first;
    };
    std::sort(bfk.begin(), bfk.end(), comp);

    bnd_descriptors.resize( num_faces() );
    size_t fbase = 0;
    /* For each entity */
    for (auto& e : entities)
    {
        /* and for each face */
        for (size_t iF = 0; iF < e.num_faces(); iF++)
        {
            /* get the element key */
            element_key fk(e.face(iF));
            auto lbcomp = [](const bfk_t& a, const element_key& b) -> bool {
                return a.first < b;
            };
            /* and lookup it in the vector we created before. */
            auto itor = std::lower_bound(bfk.begin(), bfk.end(), fk, lbcomp);
            if ( (itor == bfk.end()) or not ((*itor).first == fk) )
                continue;
            
            /* If found */
            boundary_descriptor bd;
            /* determine if it is an interface or a boundary */
            if (conn.num_neighbours(fk) == 1)
                bd.type = face_type::BOUNDARY;
            else
                bd.type = face_type::INTERFACE;
            /* and store the gmsh tag in the descriptor. */
            bd.gmsh_entity = (*itor).second;

            /* Finally, put the descriptor in the global array of faces. */
            bnd_descriptors.at(fbase + iF) = bd;
        }
        fbase += e.num_faces();
    }

    /*
    size_t normal = 0;
    size_t interface = 0;
    size_t boundary = 0;
    for (auto& bd : boundary_descriptors)
    {
        switch (bd.type)
        {
            case face_type::NONE: normal++; break;
            case face_type::INTERFACE: interface++; break;
            case face_type::BOUNDARY: boundary++; break;
        }
    }

    std::cout << bfk.size() << " " << boundary_descriptors.size() << std::endl;
    std::cout << normal << " " << interface << " " << boundary << std::endl;
    */
}
