#include "gmsh_io.h"
#include "maxwell_interface.h"
#include "maxwell_common.h"

#define DIRICHLET   2.0
#define ROBIN       1.0
#define NEUMANN     0.0

#define SEC_BCONDS              "bndconds"
#define BC_KIND                 "kind"
#define BC_SOURCE               "source"
#define BCOND_TYPE_PEC          "pec"
#define BCOND_TYPE_PMC          "pmc"
#define BCOND_TYPE_IMPEDANCE    "impedance"
#define BCOND_TYPE_EPLANEW      "plane_wave_E"
#define BCOND_TYPE_HPLANEW      "plane_wave_H"
#define BCOND_TYPE_EFIELD       "E_field"
#define BCOND_TYPE_HFIELD       "H_field"
#define BCOND_TYPE_SURFCURR     "surface_current"

#define SEC_IFCONDS             "ifaceconds"
#define IFC_KIND                "kind"
#define IFC_SOURCE              "source"
#define IFCOND_TYPE_EFIELD      "E_field"
#define IFCOND_TYPE_HFIELD      "H_field"
#define IFCOND_TYPE_SURFCURR    "surface_current"


#define SEC_MATERIALS           "materials"
#define MAT_EPS                 "epsilon"
#define MAT_EPS0                "eps0"
#define MAT_MU                  "mu"
#define MAT_MU0                 "mu0"
#define MAT_SIGMA               "sigma"

namespace maxwell {

/* Take a model and a parameter loader and return a vector with the
 * coefficients 0, 1 and 2 to select the appropriate boundary condition.
 * See the jump kernels for details about the 0, 1, and 2. */
void
init_boundary_conditions(const model& mod, const parameter_loader& mpl,
    vecxd& bcc)
{
    auto bds = mod.boundary_descriptors();
    bcc = 2*vecxd::Ones( mod.num_fluxes() );

    size_t face_num_base = 0;
    for (auto& e : mod)
    {
        for (size_t iF = 0; iF < e.num_faces(); iF++)
        {
            auto gfnum = face_num_base + iF;
            auto bd = bds[gfnum];
            if (bd.type == face_type::NONE)
                continue;

            double bc_coeff = DIRICHLET;
            switch ( mpl.boundary_type(bd.gmsh_entity) )
            {
                case boundary_condition::UNSPECIFIED:
                case boundary_condition::PEC:
                case boundary_condition::E_FIELD:
                    bc_coeff = DIRICHLET;
                    break;

                case boundary_condition::IMPEDANCE:
                case boundary_condition::PLANE_WAVE_E:
                case boundary_condition::PLANE_WAVE_H:
                    bc_coeff = ROBIN;
                    break;

                case boundary_condition::PMC:
                case boundary_condition::SURFACE_CURRENT:
                case boundary_condition::H_FIELD:
                    bc_coeff = NEUMANN;
                    break;
            }

            auto& pf = e.face(iF);
            auto& rf = e.face_refelem(pf);
            auto num_bf = rf.num_basis_functions();
            for (size_t i = 0; i < num_bf; i++)
                bcc(gfnum*num_bf + i) = bc_coeff;
        }

        face_num_base += e.num_faces();
    }
}

/* Take a model and a parameter loader and return the four vectors with
 * the Lax-Milgram flux coefficients */
void
init_lax_milgram(const model& mod, const parameter_loader& mpl,
    vecxd& paE, vecxd& pbE, vecxd& paH, vecxd& pbH)
{
    paE = vecxd::Zero( mod.num_fluxes() );
    pbE = vecxd::Zero( mod.num_fluxes() );
    paH = vecxd::Zero( mod.num_fluxes() );
    pbH = vecxd::Zero( mod.num_fluxes() );

    auto& conn = mod.connectivity();
    for (auto& e : mod)
    {
        for (size_t iT = 0; iT < e.num_cells(); iT++)
        {
            auto& pe = e.cell(iT);
            auto bar = pe.barycenter();
            auto tag = e.gmsh_tag();
            auto my_eps = mpl.epsilon(tag, bar);
            auto my_mu = mpl.mu(tag, bar);

            auto my_Z = std::sqrt(my_mu/my_eps);
            auto my_Y = 1./my_Z;

            for (size_t iF = 0; iF < e.num_faces_per_elem(); iF++)
            {
                auto face_num = e.num_faces_per_elem()*iT+iF;
                auto& pf = e.face(face_num);
                auto& rf = e.face_refelem(pf);
                auto fk = element_key(pf);

                auto glob_ofs = e.flux_base() + face_num*rf.num_basis_functions();

                auto [neigh, has_neigh] = conn.neighbour_via(e.number(), iT, fk);
                if (has_neigh)
                {
                    auto& ne_e = mod.entity_at(neigh.iE);
                    auto& ne_pe = ne_e.cell(neigh.iT);
                    auto ne_bar = ne_pe.barycenter();
                    auto ne_tag = ne_e.gmsh_tag();
                    auto ne_eps = mpl.epsilon(ne_tag, ne_bar);
                    auto ne_mu = mpl.mu(ne_tag, ne_bar);

                    auto ne_Z = std::sqrt(ne_mu/ne_eps);
                    auto ne_Y = 1./ne_Z;

                    auto aE = ne_Z/(my_Z + ne_Z);
                    auto bE = 1./(my_Z + ne_Z);
                    auto aH = ne_Y/(my_Y + ne_Y);
                    auto bH = 1./(my_Y + ne_Y);

                    for (size_t iD = 0; iD < rf.num_basis_functions(); iD++)
                    {
                        paE(glob_ofs+iD) = aE;
                        pbE(glob_ofs+iD) = bE;
                        paH(glob_ofs+iD) = aH;
                        pbH(glob_ofs+iD) = bH;
                    }
                }
                else
                {
                    for (size_t iD = 0; iD < rf.num_basis_functions(); iD++)
                    {
                        paE(glob_ofs+iD) = 0.5;
                        pbE(glob_ofs+iD) = 0.5/my_Z;
                        paH(glob_ofs+iD) = 0.5;
                        pbH(glob_ofs+iD) = 0.5/my_Y;
                    }
                }
            }
        }
    }
}


parameter_loader::parameter_loader()
{
    lua["const"][MAT_EPS0] = 8.8541878128e-12;
    lua["const"][MAT_MU0]  = 4e-7*M_PI;
}

bool
parameter_loader::validate_materials(const std::string& mpn, int tag) const
{
    auto mfun = lua[SEC_MATERIALS][mpn];
    if ( mfun.valid() )
        return true;

    auto mparams = lua[SEC_MATERIALS][tag];
    if ( mparams.valid() )
    {
        auto matparams_mpn = lua[SEC_MATERIALS][tag][mpn];
        if ( matparams_mpn.valid() )
            return true;
    }

    std::cout << "[CONFIG] '" SEC_MATERIALS ".[" << tag << "]." << mpn;
    std::cout << "' not defined and '" SEC_MATERIALS "." << mpn;
    std::cout << "(tag,x,y,z)' not present." << std::endl;
    return false;
}

bool
parameter_loader::validate_boundary_condition(const model& mod, int tag) const
{
    auto bc = lua[SEC_BCONDS][tag];
    if (not bc.valid())
        return true;

    if ( not bc[BC_KIND].valid() )
    {
        std::cout << "[CONFIG] '" BC_KIND "' not specified on interface ";
        std::cout << tag << std::endl;
        return false;
    }

    std::string bc_kind = bc[BC_KIND];

    if (bc_kind == BCOND_TYPE_PEC)
        return true;

    if (bc_kind == BCOND_TYPE_PMC)
        return true;

    if (bc_kind == BCOND_TYPE_IMPEDANCE)
        return true;

    if (bc_kind == BCOND_TYPE_EPLANEW)
    {
        if (bc[BC_SOURCE].valid())
            return true;
        
        std::cout << "[CONFIG] '" BC_SOURCE "' not specified for plane ";
        std::cout << "wave condition on surface " << tag << std::endl;
        return false;
    }

    std::cout << "[CONFIG] boundary condition not implemented on ";
    std::cout << "surface " << tag << std::endl;
    return false;
}

bool
parameter_loader::validate_interface_condition(const model& mod, int tag) const
{
    auto ic = lua[SEC_IFCONDS][tag];
    if (not ic.valid())
        return true;

    if ( not ic[IFC_KIND].valid() )
    {
        std::cout << "[CONFIG '" IFC_KIND "' not specified on interface ";
        std::cout << tag << std::endl;
        return false;
    }

    std::string ic_kind = ic[IFC_KIND];

    if (ic_kind == IFCOND_TYPE_SURFCURR)
    {
        if (ic[IFC_SOURCE].valid())
            return true;
        
        std::cout << "[CONFIG] '" IFC_SOURCE "' not specified for surface ";
        std::cout << "current condition on surface " << tag << std::endl;
        return false;
    }

    std::cout << "[CONFIG] interface condition not implemented on ";
    std::cout << "surface " << tag << std::endl;
    return false;
}

bool
parameter_loader::validate_conditions(const model& mod) const
{
    bool success = true;
    auto bds = mod.boundary_descriptors();
    std::sort(bds.begin(), bds.end());
    bds.erase( std::unique(bds.begin(), bds.end()), bds.end() );
    
    auto ft_none = [](const boundary_descriptor& bd) -> bool {
        return bd.type == face_type::NONE;
    };

    bds.erase( std::remove_if(bds.begin(), bds.end(), ft_none), bds.end() );

    for (auto& bd : bds)
    {
        switch (bd.type)
        {
            case face_type::NONE:
                assert(false);
                break;

            case face_type::BOUNDARY:
                if (not validate_boundary_condition(mod, bd.gmsh_entity) )
                    success = false;
                break;

            case face_type::INTERFACE:
                if (not validate_interface_condition(mod, bd.gmsh_entity) )
                    success = false;
                break;
        }
    }

    return success;
}

bool
parameter_loader::validate_model_params(const model& mod) const
{
    bool success = true;
    for (auto& e : mod)
    {
        auto tag = e.gmsh_tag();
        bool eps_valid = validate_materials(MAT_EPS, tag);
        bool mu_valid = validate_materials(MAT_MU, tag);
        bool sigma_valid = validate_materials(MAT_SIGMA, tag);
        if ( (not eps_valid) or (not mu_valid) or (not sigma_valid) )
            success = false;
    }

    if ( not validate_conditions(mod) )
        success = false;

    return success;
}

bool
parameter_loader::initial_Efield_defined(void) const
{
    auto eic = lua["electric_initial_condition"];
    return eic.valid();
}

vec3d
parameter_loader::initial_Efield(const point_3d& pt) const
{
    vec3d ret;
    sol::tie(ret(0), ret(1), ret(2)) =
        lua["electric_initial_condition"](pt.x(), pt.y(), pt.z());
    return ret;
}

bool
parameter_loader::initial_Hfield_defined(void) const
{
    auto mic = lua["magnetic_initial_condition"];
    return mic.valid();
}

vec3d
parameter_loader::initial_Hfield(const point_3d& pt) const
{
    vec3d ret;
    sol::tie(ret(0), ret(1), ret(2)) =
        lua["magnetic_initial_condition"](pt.x(), pt.y(), pt.z());
    return ret;
}

double
parameter_loader::epsilon(int tag, const point_3d& pt) const
{
    double eps0 = lua["const"]["eps0"];
    auto eps_dom = lua[SEC_MATERIALS][tag][MAT_EPS];
    if (eps_dom.valid())
    {
        double ret = eps_dom;
        return eps0*ret;
    }

    double ret = lua[SEC_MATERIALS][MAT_EPS](tag, pt.x(), pt.y(), pt.z());
    return eps0*ret;
}

double
parameter_loader::mu(int tag, const point_3d& pt) const
{
    double mu0 = lua["const"]["mu0"];
    auto mu_dom = lua[SEC_MATERIALS][tag][MAT_MU];
    if (mu_dom.valid())
    {
        double ret = mu_dom;
        return mu0*ret;
    }

    double ret = lua[SEC_MATERIALS][MAT_MU](tag, pt.x(), pt.y(), pt.z());
    return mu0*ret;
}

double
parameter_loader::sigma(int tag, const point_3d& pt) const
{
    auto sigma_dom = lua[SEC_MATERIALS][tag][MAT_SIGMA];
    if (sigma_dom.valid())
        return sigma_dom;

    return lua[SEC_MATERIALS][MAT_SIGMA](tag, pt.x(), pt.y(), pt.z());
}

boundary_condition
parameter_loader::boundary_type(int tag) const
{
    auto bnd_data = lua[SEC_BCONDS][tag];
    if (not bnd_data.valid())
        return boundary_condition::UNSPECIFIED;

    std::string kind_str = bnd_data[BC_KIND];

    if (kind_str == BCOND_TYPE_PEC)
        return boundary_condition::PEC;

    if (kind_str == BCOND_TYPE_PMC)
        return boundary_condition::PMC;

    if (kind_str == BCOND_TYPE_IMPEDANCE)
        return boundary_condition::IMPEDANCE;

    if (kind_str == BCOND_TYPE_EPLANEW)
        return boundary_condition::PLANE_WAVE_E;

    if (kind_str == BCOND_TYPE_HPLANEW)
        return boundary_condition::PLANE_WAVE_H;

    if (kind_str == BCOND_TYPE_EFIELD)
        return boundary_condition::E_FIELD;
    
    if (kind_str == BCOND_TYPE_HFIELD)
        return boundary_condition::H_FIELD;

    if (kind_str == BCOND_TYPE_SURFCURR)
        return boundary_condition::SURFACE_CURRENT;

    return boundary_condition::UNSPECIFIED;
}

interface_condition
parameter_loader::interface_type(int tag) const
{
    auto if_data = lua[SEC_IFCONDS][tag];
    if (not if_data.valid())
        return interface_condition::UNSPECIFIED;
    
    auto kind = if_data[IFC_KIND];

    std::string kind_str = kind;

    if (kind_str == IFCOND_TYPE_EFIELD)
        return interface_condition::E_FIELD;
    
    if (kind_str == IFCOND_TYPE_HFIELD)
        return interface_condition::H_FIELD;

    if (kind_str == IFCOND_TYPE_SURFCURR)
        return interface_condition::SURFACE_CURRENT;

    return interface_condition::UNSPECIFIED;
}

vec3d
parameter_loader::eval_boundary_source(int tag, const point_3d& pt, double t) const
{
    vec3d ret;
    sol::tie(ret(0), ret(1), ret(2)) =
        lua[SEC_BCONDS][tag][BC_SOURCE](tag, pt.x(), pt.y(), pt.z(), t);
    return ret;
}

vec3d
parameter_loader::eval_interface_source(int tag, const point_3d& pt, double t) const
{
    vec3d ret;
    sol::tie(ret(0), ret(1), ret(2)) =
        lua[SEC_IFCONDS][tag][IFC_SOURCE](tag, pt.x(), pt.y(), pt.z(), t);
    return ret;
}

} // namespace maxwell
