#include <iostream>

#include "gmsh_io.h"
#include "maxwell_interface.h"
#include "kernels_cpu.h"
#include "silo_output.hpp"
#include "timecounter.h"

namespace maxwell {

void
init_matparams(const model& mod, solver_state& state,
    const parameter_loader& mpl)
{
    state.matparams.num_dofs = mod.num_dofs();
    state.matparams.num_fluxes = mod.num_fluxes();
    state.matparams.inv_epsilon = vecxd::Zero( mod.num_dofs() );
    state.matparams.inv_mu = vecxd::Zero( mod.num_dofs() );
    state.matparams.sigma = vecxd::Zero( mod.num_dofs() );
    state.matparams.sigma_over_epsilon = vecxd::Zero( mod.num_dofs() );

    for (auto& e : mod)
    {
        auto tag = e.gmsh_tag();
        for (size_t iT = 0; iT < e.num_cells(); iT++)
        {
            auto& pe = e.cell(iT);
            auto& re = e.cell_refelem(pe);
            auto bar = pe.barycenter();
            auto epsilon = mpl.epsilon(tag, bar);
            auto mu = mpl.mu(tag, bar);
            auto sigma = mpl.sigma(tag, bar);
            auto ofs = e.cell_global_dof_offset(iT);

            for (size_t iD = 0; iD < re.num_basis_functions(); iD++)
            {
                state.matparams.inv_epsilon(ofs+iD) = 1./epsilon;
                state.matparams.inv_mu(ofs+iD) = 1./mu;
                state.matparams.sigma(ofs+iD) = sigma;
                state.matparams.sigma_over_epsilon(ofs+iD) = sigma/epsilon;
            }
        }
    }

    init_lax_milgram(mod, mpl, state.matparams.aE, state.matparams.bE,
        state.matparams.aH, state.matparams.bH);
    
    init_boundary_conditions(mod, mpl, state.matparams.bc_coeffs);
}



void
eval_interface_sources(const model& mod, const parameter_loader& mpl,
    solver_state& state, field& srcfield)
{
    auto& bds = mod.boundary_descriptors();
    assert( srcfield.num_dofs == mod.num_fluxes() );

    size_t face_num_base = 0;
    for (auto& e : mod)
    {
        for (size_t iF = 0; iF < e.num_faces(); iF++)
        {
            vec3d n = state.eds[e.number()].normals.row(iF);

            auto gfnum = face_num_base + iF;
            auto bd = bds[gfnum];
            if (bd.type == face_type::NONE or bd.type == face_type::BOUNDARY)
                continue;

            switch ( mpl.interface_type(bd.gmsh_entity) )
            {
                case interface_condition::UNSPECIFIED:
                    break;
                
                case interface_condition::E_FIELD: {
                    /* This is BS actually... */
                    auto fE = [&](const point_3d& pt) -> vec3d {
                        return mpl.eval_interface_source(bd.gmsh_entity, pt, state.curr_time);
                    };

                    matxd prjE = e.project_on_face(iF, fE);
                    auto num_bf = prjE.rows();
                    srcfield.Ex.segment(gfnum*num_bf, num_bf) = prjE.col(0);
                    srcfield.Ey.segment(gfnum*num_bf, num_bf) = prjE.col(1);
                    srcfield.Ez.segment(gfnum*num_bf, num_bf) = prjE.col(2);
                } break;

                case interface_condition::SURFACE_CURRENT: {
                    auto fJ = [&](const point_3d& pt) -> vec3d {
                        vec3d J = mpl.eval_interface_source(bd.gmsh_entity, pt, state.curr_time);
                        return n.cross(J);
                    };

                    matxd prjJ = e.project_on_face(iF, fJ);
                    auto num_bf = prjJ.rows();
                    srcfield.Hx.segment(gfnum*num_bf, num_bf) = prjJ.col(0);
                    srcfield.Hy.segment(gfnum*num_bf, num_bf) = prjJ.col(1);
                    srcfield.Hz.segment(gfnum*num_bf, num_bf) = prjJ.col(2);
                } break;

                case interface_condition::H_FIELD:
                    break;
            }
        }

        face_num_base += e.num_faces();
    }
}

void
init_from_model(const model& mod, solver_state& state)
{
    state.emf_curr.resize( mod.num_dofs() );
    state.emf_next.resize( mod.num_dofs() );
    state.dx.resize( mod.num_dofs() );
    state.dy.resize( mod.num_dofs() );
    state.dz.resize( mod.num_dofs() );
    
    state.k1.resize( mod.num_dofs() );
    state.k2.resize( mod.num_dofs() );
    state.k3.resize( mod.num_dofs() );
    state.k4.resize( mod.num_dofs() );
    state.tmp.resize( mod.num_dofs() );

    state.jumps.resize( mod.num_fluxes() );
    state.fluxes.resize( mod.num_fluxes() );

    state.bndsrcs.resize( mod.num_fluxes() );

    for (auto& e : mod)
    {
        entity_data_cpu ed;
        e.populate_entity_data(ed, mod);
        state.eds.push_back( std::move(ed) );
    }

    state.curr_time = 0.0;
    state.curr_timestep = 0;
}

static void
compute_curls(solver_state& state, const field& curr, field& next)
{
    for (const auto& ed : state.eds)
    {
        compute_field_derivatives(ed, curr.Ex, state.dx.Ex, state.dy.Ex, state.dz.Ex);
        compute_field_derivatives(ed, curr.Ey, state.dx.Ey, state.dy.Ey, state.dz.Ey);
        compute_field_derivatives(ed, curr.Ez, state.dx.Ez, state.dy.Ez, state.dz.Ez);
        compute_field_derivatives(ed, curr.Hx, state.dx.Hx, state.dy.Hx, state.dz.Hx);
        compute_field_derivatives(ed, curr.Hy, state.dx.Hy, state.dy.Hy, state.dz.Hy);
        compute_field_derivatives(ed, curr.Hz, state.dx.Hz, state.dy.Hz, state.dz.Hz);
    }

    next.Ex = state.dy.Hz - state.dz.Hy;
    next.Ey = state.dz.Hx - state.dx.Hz;
    next.Ez = state.dx.Hy - state.dy.Hx;
    next.Hx = state.dz.Ey - state.dy.Ez;
    next.Hy = state.dx.Ez - state.dz.Ex;
    next.Hz = state.dy.Ex - state.dx.Ey;
}

static void
compute_field_jumps(solver_state& state, const field& in)
{
    for (auto& ed : state.eds)
    {
        size_t num_all_fluxes = num_elems_all_orientations(ed) * ed.num_faces_per_elem * ed.num_fluxes;
 
        #pragma omp parallel for
        for (size_t i = 0; i < num_all_fluxes; i++)
        {
            auto lofs = i;
            auto gofs = ed.flux_base + i;
            if (ed.fluxdofs_neigh[lofs] != NOT_PRESENT)
            {
                auto pos_mine  = ed.fluxdofs_mine[lofs];
                auto pos_neigh = ed.fluxdofs_neigh[lofs];
                state.jumps.Ex[gofs] = in.Ex[pos_mine] - in.Ex[pos_neigh];
                state.jumps.Ey[gofs] = in.Ey[pos_mine] - in.Ey[pos_neigh];
                state.jumps.Ez[gofs] = in.Ez[pos_mine] - in.Ez[pos_neigh];
                state.jumps.Hx[gofs] = in.Hx[pos_mine] - in.Hx[pos_neigh];
                state.jumps.Hy[gofs] = in.Hy[pos_mine] - in.Hy[pos_neigh];
                state.jumps.Hz[gofs] = in.Hz[pos_mine] - in.Hz[pos_neigh];
            }
            else
            {
                auto bc = state.matparams.bc_coeffs[gofs];
                auto pos_mine  = ed.fluxdofs_mine[lofs];
                state.jumps.Ex[gofs] = bc * in.Ex[pos_mine];
                state.jumps.Ey[gofs] = bc * in.Ey[pos_mine];
                state.jumps.Ez[gofs] = bc * in.Ez[pos_mine];
                state.jumps.Hx[gofs] = (2.0 - bc) * in.Hx[pos_mine];
                state.jumps.Hy[gofs] = (2.0 - bc) * in.Hy[pos_mine];
                state.jumps.Hz[gofs] = (2.0 - bc) * in.Hz[pos_mine];
            }
        }
    }
}

static void
compute_fluxes_planar(solver_state& state)
{
    for (auto& ed : state.eds)
    {
        size_t num_all_faces = num_elems_all_orientations(ed) * ed.num_faces_per_elem;

        #pragma omp parallel for
        for (size_t iF = 0; iF < num_all_faces; iF++)
        {
            auto n = ed.normals.row(iF);
            auto nx = n[0];
            auto ny = n[1];
            auto nz = n[2];
            auto face_det = ed.face_determinants(iF);

            for (size_t i = 0; i < ed.num_fluxes; i++)
            {
                auto lofs = iF*ed.num_fluxes + i;
                auto gofs = ed.flux_base + lofs;

                /* Sources are imposed on jumps */
                auto jEx = state.jumps.Ex[gofs] - state.bndsrcs.Ex[gofs];
                auto jEy = state.jumps.Ey[gofs] - state.bndsrcs.Ey[gofs];
                auto jEz = state.jumps.Ez[gofs] - state.bndsrcs.Ez[gofs];
                auto jHx = state.jumps.Hx[gofs] + state.bndsrcs.Hx[gofs];
                auto jHy = state.jumps.Hy[gofs] + state.bndsrcs.Hy[gofs];
                auto jHz = state.jumps.Hz[gofs] + state.bndsrcs.Hz[gofs];

                auto ndotE = nx*jEx + ny*jEy + nz*jEz;
                auto ndotH = nx*jHx + ny*jHy + nz*jHz;

                auto aE = face_det * state.matparams.aE[gofs];
                auto bE = face_det * state.matparams.bE[gofs];
                auto aH = face_det * state.matparams.aH[gofs];
                auto bH = face_det * state.matparams.bH[gofs];

                /* Compute fluxes */
                state.fluxes.Ex[gofs] = aE*(nz*jHy - ny*jHz) + bE*(ndotE*nx - jEx);
                state.fluxes.Ey[gofs] = aE*(nx*jHz - nz*jHx) + bE*(ndotE*ny - jEy);
                state.fluxes.Ez[gofs] = aE*(ny*jHx - nx*jHy) + bE*(ndotE*nz - jEz);
                state.fluxes.Hx[gofs] = aH*(ny*jEz - nz*jEy) + bH*(ndotH*nx - jHx);
                state.fluxes.Hy[gofs] = aH*(nz*jEx - nx*jEz) + bH*(ndotH*ny - jHy);
                state.fluxes.Hz[gofs] = aH*(nx*jEy - ny*jEx) + bH*(ndotH*nz - jHz);
            }
        }
    }
}

static void
compute_fluxes(solver_state& state, const field& in, field& out)
{
    compute_field_jumps(state, in);
    compute_fluxes_planar(state);
    
    for (auto& ed : state.eds)
    {
        compute_flux_lifting(ed, state.fluxes.Ex, out.Ex);
        compute_flux_lifting(ed, state.fluxes.Ey, out.Ey);
        compute_flux_lifting(ed, state.fluxes.Ez, out.Ez);
        compute_flux_lifting(ed, state.fluxes.Hx, out.Hx);
        compute_flux_lifting(ed, state.fluxes.Hy, out.Hy);
        compute_flux_lifting(ed, state.fluxes.Hz, out.Hz);
    }
}

static void
compute_euler_update(solver_state& state, const field& y, const field& k, double dt, field& out)
{
    #pragma omp parallel for
    for (size_t i = 0; i < out.num_dofs; i++)
    {
        auto CR = 1 - dt*state.matparams.sigma_over_epsilon[i];
        auto CC = dt*state.matparams.inv_epsilon[i];
        out.Ex[i] = y.Ex[i]*CR + k.Ex[i]*CC;
        out.Ey[i] = y.Ey[i]*CR + k.Ey[i]*CC;
        out.Ez[i] = y.Ez[i]*CR + k.Ez[i]*CC;
    }
    
    #pragma omp parallel for
    for (size_t i = 0; i < out.num_dofs; i++)
    {
        auto CC = dt*state.matparams.inv_mu[i];
        out.Hx[i] = y.Hx[i] + k.Hx[i]*CC;
        out.Hy[i] = y.Hy[i] + k.Hy[i]*CC;
        out.Hz[i] = y.Hz[i] + k.Hz[i]*CC;
    }
}

static void
compute_rk4_weighted_sum(solver_state& state, const field& in, double dt, field& out)
{
    #pragma omp parallel for
    for (size_t i = 0; i < out.num_dofs; i++)
    {
        auto CR = 1 - dt*state.matparams.sigma_over_epsilon[i];
        auto CC = dt*state.matparams.inv_epsilon[i];
        out.Ex[i] = in.Ex[i]*CR + CC*(state.k1.Ex[i] + 2*state.k2.Ex[i] + 2*state.k3.Ex[i] + state.k4.Ex[i])/6;
        out.Ey[i] = in.Ey[i]*CR + CC*(state.k1.Ey[i] + 2*state.k2.Ey[i] + 2*state.k3.Ey[i] + state.k4.Ey[i])/6;
        out.Ez[i] = in.Ez[i]*CR + CC*(state.k1.Ez[i] + 2*state.k2.Ez[i] + 2*state.k3.Ez[i] + state.k4.Ez[i])/6;
    }

    #pragma omp parallel for
    for (size_t i = 0; i < out.num_dofs; i++)
    {
        auto CC = dt*state.matparams.inv_mu[i];
        out.Hx[i] = in.Hx[i] + CC*(state.k1.Hx[i] + 2*state.k2.Hx[i] + 2*state.k3.Hx[i] + state.k4.Hx[i])/6;
        out.Hy[i] = in.Hy[i] + CC*(state.k1.Hy[i] + 2*state.k2.Hy[i] + 2*state.k3.Hy[i] + state.k4.Hy[i])/6;
        out.Hz[i] = in.Hz[i] + CC*(state.k1.Hz[i] + 2*state.k2.Hz[i] + 2*state.k3.Hz[i] + state.k4.Hz[i])/6;
    }
}

static void
apply_operator(solver_state& state, const field& curr, field& next)
{
    compute_curls(state, curr, next);
    compute_fluxes(state, curr, next);
}

void
timestep(solver_state& state)
{
    /*
    apply_operator(state, state.emf_curr, state.tmp);
    compute_euler_update(state, state.emf_curr, state.tmp, state.delta_t, state.emf_next);
    */

    apply_operator(state, state.emf_curr, state.k1);

    compute_euler_update(state, state.emf_curr, state.k1, 0.5*state.delta_t, state.tmp);
    apply_operator(state, state.tmp, state.k2);

    compute_euler_update(state, state.emf_curr, state.k2, 0.5*state.delta_t, state.tmp);
    apply_operator(state, state.tmp, state.k3);
        
    compute_euler_update(state, state.emf_curr, state.k3, state.delta_t, state.tmp);
    apply_operator(state, state.tmp, state.k4);

    compute_rk4_weighted_sum(state, state.emf_curr, state.delta_t, state.emf_next);

    state.curr_time += state.delta_t;
    state.curr_timestep += 1;
}

void
export_to_silo(const model& mod, const solver_state& state, const std::string& filename)
{
    silo sdb;
    sdb.create_db(filename);
    sdb.import_mesh_from_gmsh();
    sdb.write_mesh(state.curr_time, state.curr_timestep);

    std::vector<double> curr_Ex; curr_Ex.resize( mod.num_cells() );
    std::vector<double> curr_Ey; curr_Ey.resize( mod.num_cells() );
    std::vector<double> curr_Ez; curr_Ez.resize( mod.num_cells() );
    std::vector<double> curr_Hx; curr_Hx.resize( mod.num_cells() );
    std::vector<double> curr_Hy; curr_Hy.resize( mod.num_cells() );
    std::vector<double> curr_Hz; curr_Hz.resize( mod.num_cells() );

    std::vector<double> next_Ex; next_Ex.resize( mod.num_cells() );
    std::vector<double> next_Ey; next_Ey.resize( mod.num_cells() );
    std::vector<double> next_Ez; next_Ez.resize( mod.num_cells() );
    std::vector<double> next_Hx; next_Hx.resize( mod.num_cells() );
    std::vector<double> next_Hy; next_Hy.resize( mod.num_cells() );
    std::vector<double> next_Hz; next_Hz.resize( mod.num_cells() );

    for (size_t iE = 0; iE < mod.num_entities(); iE++)
    {
        auto& e = mod.entity_at(iE);
        for (size_t iT = 0; iT < e.num_cells(); iT++)
        {
            auto& pe = e.cell(iT);
            auto& re = e.cell_refelem(pe);
            vecxd phi_bar = re.basis_functions({1./3., 1./3., 1./3.});
            auto num_bf = re.num_basis_functions();
            auto ofs = e.cell_global_dof_offset(iT);
            auto gi = e.cell_global_index_by_gmsh(iT);
            curr_Ex[gi] = state.emf_curr.Ex.segment(ofs, num_bf).dot(phi_bar);
            curr_Ey[gi] = state.emf_curr.Ey.segment(ofs, num_bf).dot(phi_bar);
            curr_Ez[gi] = state.emf_curr.Ez.segment(ofs, num_bf).dot(phi_bar);
            curr_Hx[gi] = state.emf_curr.Hx.segment(ofs, num_bf).dot(phi_bar);
            curr_Hy[gi] = state.emf_curr.Hy.segment(ofs, num_bf).dot(phi_bar);
            curr_Hz[gi] = state.emf_curr.Hz.segment(ofs, num_bf).dot(phi_bar);
            next_Ex[gi] = state.emf_next.Ex.segment(ofs, num_bf).dot(phi_bar);
            next_Ey[gi] = state.emf_next.Ey.segment(ofs, num_bf).dot(phi_bar);
            next_Ez[gi] = state.emf_next.Ez.segment(ofs, num_bf).dot(phi_bar);
            next_Hx[gi] = state.emf_next.Hx.segment(ofs, num_bf).dot(phi_bar);
            next_Hy[gi] = state.emf_next.Hy.segment(ofs, num_bf).dot(phi_bar);
            next_Hz[gi] = state.emf_next.Hz.segment(ofs, num_bf).dot(phi_bar);
        }
    }

    sdb.write_zonal_variable("curr_Ex", curr_Ex);
    sdb.write_zonal_variable("curr_Ey", curr_Ey);
    sdb.write_zonal_variable("curr_Ez", curr_Ez);
    sdb.write_zonal_variable("curr_Hx", curr_Hx);
    sdb.write_zonal_variable("curr_Hy", curr_Hy);
    sdb.write_zonal_variable("curr_Hz", curr_Hz);
    sdb.write_vector_expression("curr_E", "{curr_Ex, curr_Ey, curr_Ez}");
    sdb.write_vector_expression("curr_H", "{curr_Hx, curr_Hy, curr_Hz}");
    sdb.write_zonal_variable("next_Ex", next_Ex);
    sdb.write_zonal_variable("next_Ey", next_Ey);
    sdb.write_zonal_variable("next_Ez", next_Ez);
    sdb.write_zonal_variable("next_Hx", next_Hx);
    sdb.write_zonal_variable("next_Hy", next_Hy);
    sdb.write_zonal_variable("next_Hz", next_Hz);
    sdb.write_vector_expression("next_E", "{next_Ex, next_Ey, next_Ez}");
    sdb.write_vector_expression("next_H", "{next_Hx, next_Hy, next_Hz}");
}

} // namespace maxwell
