#include <iostream>

#include "gmsh_io.h"
#include "maxwell_interface.h"
#include "kernels_cpu.h"
#include "kernels_gpu.h"
#include "silo_output.hpp"
#include "timecounter.h"


namespace maxwell {

void
init_matparams(const model& mod, solver_state_gpu& state,
    const parameter_loader& mpl)
{
    vecxd inv_eps = vecxd::Zero( mod.num_dofs() );
    vecxd inv_mu = vecxd::Zero( mod.num_dofs() );
    vecxd sigma = vecxd::Zero( mod.num_dofs() );
    vecxd sigma_over_eps = vecxd::Zero( mod.num_dofs() );

    vecxd Z = vecxd::Zero( mod.num_dofs() );
    vecxd Y = vecxd::Zero( mod.num_dofs() );

    for (auto& e : mod)
    {
        auto tag = e.gmsh_tag();
        for (size_t iT = 0; iT < e.num_cells(); iT++)
        {
            auto& pe = e.cell(iT);
            auto& re = e.cell_refelem(pe);
            auto bar = pe.barycenter();
            auto v_epsilon = mpl.epsilon(tag, bar);
            auto v_mu = mpl.mu(tag, bar);
            auto v_sigma = mpl.sigma(tag, bar);
            auto ofs = e.cell_global_dof_offset(iT);

            for (size_t iD = 0; iD < re.num_basis_functions(); iD++)
            {
                inv_eps(ofs+iD) = 1./v_epsilon;
                inv_mu(ofs+iD) = 1./v_mu;
                sigma(ofs+iD) = v_sigma;
                sigma_over_eps(ofs+iD) = v_sigma/v_epsilon;
            }
        }
    }

    state.matparams.inv_epsilon.copyin( inv_eps.data(), inv_eps.size() );
    state.matparams.inv_mu.copyin( inv_mu.data(), inv_mu.size() );
    state.matparams.sigma.copyin( sigma.data(), sigma.size() );
    state.matparams.sigma_over_epsilon.copyin( sigma_over_eps.data(), sigma_over_eps.size() );

    vecxd aE, bE, aH, bH;
    init_lax_milgram(mod, mpl, aE, bE, aH, bH);
    state.matparams.aE.copyin(aE.data(), aE.size());
    state.matparams.bE.copyin(bE.data(), bE.size());
    state.matparams.aH.copyin(aH.data(), aH.size());
    state.matparams.bH.copyin(bH.data(), bH.size());

    vecxd bcc;
    init_boundary_conditions(mod, mpl, bcc);
    state.matparams.bc_coeffs.copyin(bcc.data(), bcc.size());
}

void init_from_model(const model& mod, solver_state_gpu& state)
{
    state.emf_curr.resize( mod.num_dofs() );
    state.emf_next.resize( mod.num_dofs() );
    state.dx.resize( mod.num_dofs() );
    state.dy.resize( mod.num_dofs() );
    state.dz.resize( mod.num_dofs() );

    state.jumps.resize( mod.num_fluxes() );
    state.fluxes.resize( mod.num_fluxes() );

    state.k1.resize( mod.num_dofs() );
    state.k2.resize( mod.num_dofs() );
    state.k3.resize( mod.num_dofs() );
    state.k4.resize( mod.num_dofs() );
    state.tmp.resize( mod.num_dofs() );

    state.bndsrcs.resize( mod.num_fluxes() );
    state.bndsrcs_buf.resize( mod.num_fluxes() );
    state.bndsrcs_cpu.resize( mod.num_fluxes() );

    for (auto& e : mod)
    {
        entity_data_cpu ed;
        e.populate_entity_data(ed, mod);
        entity_data_gpu edg(ed);
        state.eds.push_back( std::move(ed) );
        state.edgs.push_back( std::move(edg) );
    }

    state.curr_time = 0.0;
    state.curr_timestep = 0;
}


void compute_curls(solver_state_gpu& state, const field_gpu& curr, field_gpu& next)
{
    auto currp = curr.data();
    auto nextp = next.data();
    auto dxp = state.dx.data();
    auto dyp = state.dy.data();
    auto dzp = state.dz.data();

    for (const auto& edg : state.edgs)
    {
        gpu_compute_field_derivatives(edg, currp.Ex, dxp.Ex, dyp.Ex, dzp.Ex, 1.0, state.compute_stream);
        gpu_compute_field_derivatives(edg, currp.Ey, dxp.Ey, dyp.Ey, dzp.Ey, 1.0, state.compute_stream);
        gpu_compute_field_derivatives(edg, currp.Ez, dxp.Ez, dyp.Ez, dzp.Ez, 1.0, state.compute_stream);
        gpu_compute_field_derivatives(edg, currp.Hx, dxp.Hx, dyp.Hx, dzp.Hx, 1.0, state.compute_stream);
        gpu_compute_field_derivatives(edg, currp.Hy, dxp.Hy, dyp.Hy, dzp.Hy, 1.0, state.compute_stream);
        gpu_compute_field_derivatives(edg, currp.Hz, dxp.Hz, dyp.Hz, dzp.Hz, 1.0, state.compute_stream);
    }

    
    gpu_subtract(nextp.Ex, dyp.Hz, dzp.Hy, nextp.num_dofs, state.compute_stream);
    gpu_subtract(nextp.Ey, dzp.Hx, dxp.Hz, nextp.num_dofs, state.compute_stream);
    gpu_subtract(nextp.Ez, dxp.Hy, dyp.Hx, nextp.num_dofs, state.compute_stream);
    gpu_subtract(nextp.Hx, dzp.Ey, dyp.Ez, nextp.num_dofs, state.compute_stream);
    gpu_subtract(nextp.Hy, dxp.Ez, dzp.Ex, nextp.num_dofs, state.compute_stream);
    gpu_subtract(nextp.Hz, dyp.Ex, dxp.Ey, nextp.num_dofs, state.compute_stream);
}

void
compute_fluxes(solver_state_gpu& state, const field_gpu& in, field_gpu& out)
{
    auto matparams_ptrs = state.matparams.data();
    double *bc_coeffs = matparams_ptrs.bc_coeffs;

    for (const auto& edg : state.edgs)
        maxwell::gpu_compute_jumps(edg, in, state.jumps, bc_coeffs, state.compute_stream);

    for (const auto& edg : state.edgs)
        maxwell::gpu_compute_fluxes(edg, state.jumps, state.fluxes, state.bndsrcs,
            state.matparams, state.compute_stream);

    auto fluxesp = state.fluxes.data();
    auto outp = out.data();

    for (auto& edg : state.edgs)
    {
        gpu_compute_flux_lifting(edg, fluxesp.Ex, outp.Ex, state.compute_stream);
        gpu_compute_flux_lifting(edg, fluxesp.Ey, outp.Ey, state.compute_stream);
        gpu_compute_flux_lifting(edg, fluxesp.Ez, outp.Ez, state.compute_stream);
        gpu_compute_flux_lifting(edg, fluxesp.Hx, outp.Hx, state.compute_stream);
        gpu_compute_flux_lifting(edg, fluxesp.Hy, outp.Hy, state.compute_stream);
        gpu_compute_flux_lifting(edg, fluxesp.Hz, outp.Hz, state.compute_stream);
    }
}

void
apply_operator(solver_state_gpu& state, const field_gpu& curr, field_gpu& next)
{
    compute_curls(state, curr, next);
    compute_fluxes(state, curr, next);
}

void
compute_euler_update(solver_state_gpu& state, const field_gpu& y,
    const field_gpu& k, double dt, field_gpu& out)
{
    double *r = state.matparams.sigma_over_epsilon.data();
    double *eps = state.matparams.inv_epsilon.data();
    double *mu = state.matparams.inv_mu.data();

    auto outp = out.data();
    auto yp = y.data();
    auto kp = k.data();

    gpu_compute_euler_update(outp.Ex, yp.Ex, kp.Ex, r, eps, out.num_dofs, dt, state.compute_stream);
    gpu_compute_euler_update(outp.Ey, yp.Ey, kp.Ey, r, eps, out.num_dofs, dt, state.compute_stream);
    gpu_compute_euler_update(outp.Ez, yp.Ez, kp.Ez, r, eps, out.num_dofs, dt, state.compute_stream);
    gpu_compute_euler_update(outp.Hx, yp.Hx, kp.Hx, mu, out.num_dofs, dt, state.compute_stream);
    gpu_compute_euler_update(outp.Hy, yp.Hy, kp.Hy, mu, out.num_dofs, dt, state.compute_stream);
    gpu_compute_euler_update(outp.Hz, yp.Hz, kp.Hz, mu, out.num_dofs, dt, state.compute_stream);
}

void
compute_rk4_weighted_sum(solver_state_gpu& state, const field_gpu& in,
    double dt, field_gpu& out)
{
    double *r = state.matparams.sigma_over_epsilon.data();
    double *eps = state.matparams.inv_epsilon.data();
    double *mu = state.matparams.inv_mu.data();

    auto inp = in.data();
    auto outp = out.data();
    auto k1p = state.k1.data();
    auto k2p = state.k2.data();
    auto k3p = state.k3.data();
    auto k4p = state.k4.data();

    gpu_compute_rk4_weighted_sum(outp.Ex, inp.Ex, k1p.Ex, k2p.Ex, k3p.Ex, k4p.Ex, r, eps, out.num_dofs, dt, state.compute_stream);
    gpu_compute_rk4_weighted_sum(outp.Ey, inp.Ey, k1p.Ey, k2p.Ey, k3p.Ey, k4p.Ey, r, eps, out.num_dofs, dt, state.compute_stream);
    gpu_compute_rk4_weighted_sum(outp.Ez, inp.Ez, k1p.Ez, k2p.Ez, k3p.Ez, k4p.Ez, r, eps, out.num_dofs, dt, state.compute_stream);
    gpu_compute_rk4_weighted_sum(outp.Hx, inp.Hx, k1p.Hx, k2p.Hx, k3p.Hx, k4p.Hx, mu, out.num_dofs, dt, state.compute_stream);
    gpu_compute_rk4_weighted_sum(outp.Hy, inp.Hy, k1p.Hy, k2p.Hy, k3p.Hy, k4p.Hy, mu, out.num_dofs, dt, state.compute_stream);
    gpu_compute_rk4_weighted_sum(outp.Hz, inp.Hz, k1p.Hz, k2p.Hz, k3p.Hz, k4p.Hz, mu, out.num_dofs, dt, state.compute_stream);
}

void
timestep(solver_state_gpu& state)
{
    //timecounter tc;
    //tc.tic();

    /*
    apply_operator(state, state.emf_curr, state.tmp);
    compute_euler_update(state, state.emf_curr, state.tmp, state.delta_t, state.emf_next);
    */

    apply_operator(state, state.emf_curr, state.k1);

    compute_euler_update(state, state.emf_curr, state.k1, 0.5*state.delta_t, state.tmp);
    apply_operator(state, state.tmp, state.k2);

    compute_euler_update(state, state.emf_curr, state.k2, 0.5*state.delta_t, state.tmp);
    apply_operator(state, state.tmp, state.k3);
        
    compute_euler_update(state, state.emf_curr, state.k3, state.delta_t, state.tmp);
    apply_operator(state, state.tmp, state.k4);

    compute_rk4_weighted_sum(state, state.emf_curr, state.delta_t, state.emf_next);

    state.curr_time += state.delta_t;
    state.curr_timestep += 1;

    //double time = tc.toc();
    //double dofs_per_sec = 6*state.emf_curr.num_dofs/time;
    //std::cout << "Timestep " << state.curr_timestep << ", " << dofs_per_sec << " DOFs/s" << std::endl;
}


void
export_to_silo(const model& mod, const solver_state_gpu& state, const std::string& filename)
{
    silo sdb;
    sdb.create_db(filename);
    sdb.import_mesh_from_gmsh();
    sdb.write_mesh(state.curr_time, state.curr_timestep);

    std::vector<double> curr_Ex; curr_Ex.resize( mod.num_cells() );
    std::vector<double> curr_Ey; curr_Ey.resize( mod.num_cells() );
    std::vector<double> curr_Ez; curr_Ez.resize( mod.num_cells() );
    std::vector<double> curr_Hx; curr_Hx.resize( mod.num_cells() );
    std::vector<double> curr_Hy; curr_Hy.resize( mod.num_cells() );
    std::vector<double> curr_Hz; curr_Hz.resize( mod.num_cells() );

    std::vector<double> next_Ex; next_Ex.resize( mod.num_cells() );
    std::vector<double> next_Ey; next_Ey.resize( mod.num_cells() );
    std::vector<double> next_Ez; next_Ez.resize( mod.num_cells() );
    std::vector<double> next_Hx; next_Hx.resize( mod.num_cells() );
    std::vector<double> next_Hy; next_Hy.resize( mod.num_cells() );
    std::vector<double> next_Hz; next_Hz.resize( mod.num_cells() );

    field emf_curr;
    emf_curr.resize( state.emf_curr.num_dofs );
    state.emf_curr.copyout(emf_curr);

    field emf_next;
    emf_next.resize( state.emf_next.num_dofs );
    state.emf_next.copyout(emf_next);

    for (size_t iE = 0; iE < mod.num_entities(); iE++)
    {
        auto& e = mod.entity_at(iE);
        for (size_t iT = 0; iT < e.num_cells(); iT++)
        {
            auto& pe = e.cell(iT);
            auto& re = e.cell_refelem(pe);
            vecxd phi_bar = re.basis_functions({1./3., 1./3., 1./3.});
            auto num_bf = re.num_basis_functions();
            auto ofs = e.cell_global_dof_offset(iT);
            auto gi = e.cell_global_index_by_gmsh(iT);
            curr_Ex[gi] = emf_curr.Ex.segment(ofs, num_bf).dot(phi_bar);
            curr_Ey[gi] = emf_curr.Ey.segment(ofs, num_bf).dot(phi_bar);
            curr_Ez[gi] = emf_curr.Ez.segment(ofs, num_bf).dot(phi_bar);
            curr_Hx[gi] = emf_curr.Hx.segment(ofs, num_bf).dot(phi_bar);
            curr_Hy[gi] = emf_curr.Hy.segment(ofs, num_bf).dot(phi_bar);
            curr_Hz[gi] = emf_curr.Hz.segment(ofs, num_bf).dot(phi_bar);
            next_Ex[gi] = emf_next.Ex.segment(ofs, num_bf).dot(phi_bar);
            next_Ey[gi] = emf_next.Ey.segment(ofs, num_bf).dot(phi_bar);
            next_Ez[gi] = emf_next.Ez.segment(ofs, num_bf).dot(phi_bar);
            next_Hx[gi] = emf_next.Hx.segment(ofs, num_bf).dot(phi_bar);
            next_Hy[gi] = emf_next.Hy.segment(ofs, num_bf).dot(phi_bar);
            next_Hz[gi] = emf_next.Hz.segment(ofs, num_bf).dot(phi_bar);
        }
    }

    sdb.write_zonal_variable("curr_Ex", curr_Ex);
    sdb.write_zonal_variable("curr_Ey", curr_Ey);
    sdb.write_zonal_variable("curr_Ez", curr_Ez);
    sdb.write_zonal_variable("curr_Hx", curr_Hx);
    sdb.write_zonal_variable("curr_Hy", curr_Hy);
    sdb.write_zonal_variable("curr_Hz", curr_Hz);
    sdb.write_vector_expression("curr_E", "{curr_Ex, curr_Ey, curr_Ez}");
    sdb.write_vector_expression("curr_H", "{curr_Hx, curr_Hy, curr_Hz}");
    sdb.write_zonal_variable("next_Ex", next_Ex);
    sdb.write_zonal_variable("next_Ey", next_Ey);
    sdb.write_zonal_variable("next_Ez", next_Ez);
    sdb.write_zonal_variable("next_Hx", next_Hx);
    sdb.write_zonal_variable("next_Hy", next_Hy);
    sdb.write_zonal_variable("next_Hz", next_Hz);
    sdb.write_vector_expression("next_E", "{next_Ex, next_Ey, next_Ez}");
    sdb.write_vector_expression("next_H", "{next_Hx, next_Hy, next_Hz}");
}

} // namespace maxwell
